@@ -35,9 +35,9 @@
 #
 # Following modifiers are available for "WRAP" and "GATE":
 #   - "map(varname)" will map function's parameter "varname" using
-#     the SBOX_MAP_PATH macro
+#     the sbox_map_path() function
 #   - "map_at(fdname,varname)" will map function's parameter "varname" using
-#     the SBOX_MAP_PATH_AT macro
+#     the sbox_map_path_at() function
 #   - "hardcode_param(N,name)" will hardcode name of the Nth parameter
 #     to "name" (this is typically needed only if the function definition uses
 #     macros to build the parameter list, instead of specifying names of
@@ -54,8 +54,18 @@
 #     provided check). "varname" must be the same name which was specified
 #     to map() or map_at(). "error_code" will be assigned to errno, and
 #     the failure will always be logged (SB_LOG_NOTICE level)
-#
-# For "WRAP" only:
+#   - "dont_resolve_final_symlink" is used to prefix "map" modifiers where the
+#     final symbolic link should not be followed, because the call operates
+#     on the symlink itself (for example, see the difference between stat() and
+#     lstat()). NOTE: THIS MUST BE USED BEFORE THE map() OR map_at() MODIFIERS!
+#   - "dont_resolve_final_symlink_if(condition)" conditionally
+#     leaves the final symlink unresolved (see "dont_resolve_final_symlink")
+#     NOTE: THIS MUST BE USED BEFORE THE map() OR map_at() MODIFIERS!
+#   - "resolve_final_symlink" is the opposite of "dont_resolve_final_symlink"
+#     (and it is on by default)
+#   - "postprocess(varname)" can be used to call  postprocessor functions for
+#     mapped variables.
+#   - "return(expr)" can be used to alter the return value.
 #   - "create_nomap_nolog_version" creates a direct interface function to the
 #     next function (for internal use inside the preload library)
 # For "GATE" only:
@@ -86,6 +96,16 @@ my $num_errors = 0;
 # "LOGLEVEL" was not in interface.master and b) tries to be informative
 my $generated_code_loglevel = "LOGLEVEL_statement_missing_from_interface_master";
 
+#============================================
+# This will be added to all generated interface functions:
+# (global variables need to be initialized by a function call
+# because the library constructor function seems to be unreliable:
+# it may not be the first executed function in a multithreaded
+# environment!)
+my $common_initcode_for_all_functions =
+	"\tif (!sb2_global_vars_initialized__)\n".
+	"\t\tsb2_initialize_global_variables();\n";
+
 #============================================
 
 sub write_output_file {
@@ -358,7 +378,7 @@ sub create_code_for_va_list_get_mode {
 	if($condition ne '') {
 		$condition = "if($condition) ";
 	}
-	return( "\t${condition}{\n".
+	return( "\t" . $condition . "{\n".
 		"\t\tva_list arg;\n".
                 "\t\tva_start (arg, $last_named_var);\n".
                 "\t\tmode = va_arg (arg, int);\n".
@@ -366,6 +386,17 @@ sub create_code_for_va_list_get_mode {
 		"\t}\n");
 }
 
+sub create_code_for_va_list_get_void_ptr {
+	my $last_named_var = shift;
+
+	return( "\t{\n".
+		"\t\tva_list arg;\n".
+                "\t\tva_start (arg, $last_named_var);\n".
+                "\t\topt_arg = va_arg (arg, void*);\n".
+                "\t\tva_end (arg);\n".
+		"\t}\n");
+}
+
 sub process_readonly_check_modifier {
 	my $mods = shift;
 	my $extra_check = shift;
@@ -373,8 +404,14 @@ sub process_readonly_check_modifier {
 	my $return_value = shift;
 	my $error_code = shift;
 
-	my $new_name = "mapped__".$param_to_be_mapped;
-	my $ro_flag = $param_to_be_mapped."_is_readonly";
+	my $new_name = "res_mapped__".$param_to_be_mapped.".mres_result_path";
+	my $ro_flag = "res_mapped__".$param_to_be_mapped.".mres_readonly";
+
+	if (!defined($mods->{'mapping_results_by_orig_name'}->{$param_to_be_mapped})) {
+		printf "ERROR: mapping_results_by_orig_name not found for '%s'\n",
+			$param_to_be_mapped;
+		$num_errors++;
+	}
 
 	if (defined($extra_check)) {
 		$extra_check = " && ($extra_check)";
@@ -428,10 +465,16 @@ sub process_wrap_or_gate_modifiers {
 		'va_list_handler_code' => "",
 		'va_list_end_code' => "",
 		'mapped_params_by_orig_name' => {},
+		'mapping_results_by_orig_name' => {},
+		'dont_resolve_final_symlink' => 0,
+
+		'postprocess_vars' => [],
+		'return_expr' => undef,
 
 		# processing modifiers may change the parameter list
 		# (but always we'll start with a copy of the original names)
 		'parameter_names' => [@{$fn->{'parameter_names'}}],
+		'parameter_types' => [@{$fn->{'parameter_types'}}],
 
 		'make_nomap_function' => 0,		# flag
 		'make_nomap_nolog_function' => 0,	# flag
@@ -442,6 +485,7 @@ sub process_wrap_or_gate_modifiers {
 	};
 
 	my $r_param_names = $mods->{'parameter_names'};
+	my $r_param_types = $mods->{'parameter_types'};
 
 	my $varargs_handled = 0;
 
@@ -452,37 +496,60 @@ sub process_wrap_or_gate_modifiers {
 			my $param_to_be_mapped = $1;
 
 			my $new_name = "mapped__".$param_to_be_mapped;
-			my $ro_flag = $param_to_be_mapped."_is_readonly";
+			my $no_symlink_resolve = $mods->{'dont_resolve_final_symlink'};
 
-			$mods->{'mapped_params_by_orig_name'}->{$param_to_be_mapped} = $new_name;
+			$mods->{'mapped_params_by_orig_name'}->{$param_to_be_mapped} = "res_$new_name.mres_result_path";
+			$mods->{'mapping_results_by_orig_name'}->{$param_to_be_mapped} = "res_$new_name";
 			$mods->{'path_mapping_vars'} .= 
-				"\tchar *$new_name = NULL;\n".
-				"\tint $ro_flag = 0;\n";
+				"\tmapping_results_t res_$new_name;\n";
 			$mods->{'path_mapping_code'} .=
-				"\tSBOX_MAP_PATH($param_to_be_mapped, ".
-					"$new_name, &$ro_flag);\n";
+				"\tclear_mapping_results_struct(&res_$new_name);\n".
+				"\tsbox_map_path(__func__, ".
+					"$param_to_be_mapped, ".
+					"$no_symlink_resolve, ".
+					"&res_$new_name);\n";
 			$mods->{'free_path_mapping_vars_code'} .=
-				"\tif($new_name) free($new_name);\n";
+				"\tfree_mapping_results(&res_$new_name);\n";
 
 			# Make a "..._nomap" version, because the main
 			# wrapper has mappings.
 			$mods->{'make_nomap_function'} = 1;
+		} elsif($modifiers[$i] =~ m/^postprocess\((.*)\)$/) {
+			my $param_to_postprocess = $1;
+
+			if (defined($mods->{'mapped_params_by_orig_name'}->{$param_to_postprocess}) ||
+			    ($param_to_postprocess eq '')) {
+				push(@{$mods->{'postprocess_vars'}},
+					$param_to_postprocess);
+			} else {
+				printf "ERROR: can't postprocess ".
+					"%s (parameter not mapped)\n",
+					$param_to_postprocess;
+				$num_errors++;
+			}
+		} elsif($modifiers[$i] =~ m/^return\((.*)\)$/) {
+			$mods->{'return_expr'} = $1;
 		} elsif($modifiers[$i] =~ m/^map_at\((.*),(.*)\)$/) {
 			my $fd_param = $1;
 			my $param_to_be_mapped = $2;
 
 			my $new_name = "mapped__".$param_to_be_mapped;
 			my $ro_flag = $param_to_be_mapped."_is_readonly";
+			my $no_symlink_resolve = $mods->{'dont_resolve_final_symlink'};
 
-			$mods->{'mapped_params_by_orig_name'}->{$param_to_be_mapped} = $new_name;
+			$mods->{'mapped_params_by_orig_name'}->{$param_to_be_mapped} = "res_$new_name.mres_result_path";
+			$mods->{'mapping_results_by_orig_name'}->{$param_to_be_mapped} = "res_$new_name";
 			$mods->{'path_mapping_vars'} .= 
-				"\tchar *$new_name = NULL;\n".
-				"\tint $ro_flag = 0;\n";
+				"\tmapping_results_t res_$new_name;\n";
 			$mods->{'path_mapping_code'} .=
-				"\tSBOX_MAP_PATH_AT($fd_param, ".
-				"$param_to_be_mapped, $new_name, &$ro_flag);\n";
+				"\tclear_mapping_results_struct(&res_$new_name);\n".
+				"\tsbox_map_path_at(__func__, ".
+					"$fd_param, ".
+					"$param_to_be_mapped, ".
+					"$no_symlink_resolve, ".
+					"&res_$new_name);\n";
 			$mods->{'free_path_mapping_vars_code'} .=
-				"\tif($new_name) free($new_name);\n";
+				"\tfree_mapping_results(&res_$new_name);\n";
 
 			# Make a "..._nomap" version, because the main
 			# wrapper has mappings.
@@ -504,9 +571,7 @@ sub process_wrap_or_gate_modifiers {
 			process_readonly_check_modifier($mods, $extra_check,
 				$param_to_be_mapped, $return_value, 
 				$error_code);
-		} elsif(($modifiers[$i] eq 'create_nomap_nolog_version') &&
-			($command eq 'WRAP')) {
-
+		} elsif($modifiers[$i] eq 'create_nomap_nolog_version') {
 			$mods->{'make_nomap_nolog_function'} = 1;
 		} elsif($modifiers[$i] =~ m/^hardcode_param\((.*),(.*)\)$/) {
 			my $param_number = $1;
@@ -515,16 +580,23 @@ sub process_wrap_or_gate_modifiers {
 		} elsif(($modifiers[$i] eq 'optional_arg_is_create_mode') &&
 			($fn->{'has_varargs'})) {
 			$r_param_names->[$varargs_index] = "mode";
+			$r_param_types->[$varargs_index] = "int";
 			$mods->{'local_vars_for_varargs_handler'} .=
 				"\tint mode = 0;\n";
 			$mods->{'va_list_handler_code'} =
 				create_code_for_va_list_get_mode(
 					"", $fn->{'last_named_var'});
 			$varargs_handled = 1;
+		} elsif($modifiers[$i] eq 'dont_resolve_final_symlink') {
+			$mods->{'dont_resolve_final_symlink'} = 1;
+		} elsif($modifiers[$i] =~ m/^dont_resolve_final_symlink_if\((.*)\)$/) {
+			my $condition = $1;
+			$mods->{'dont_resolve_final_symlink'} = "($condition)";
 		} elsif(($modifiers[$i] =~ m/^optional_arg_is_create_mode\((.*)\)$/) &&
 			($fn->{'has_varargs'})) {
 			my $va_list_condition = $1;
 			$r_param_names->[$varargs_index] = "mode";
+			$r_param_types->[$varargs_index] = "int";
 			$mods->{'local_vars_for_varargs_handler'} .=
 				"\tint mode = 0;\n";
 			$mods->{'va_list_handler_code'} =
@@ -532,11 +604,22 @@ sub process_wrap_or_gate_modifiers {
 					$va_list_condition,
 					$fn->{'last_named_var'});
 			$varargs_handled = 1;
+		} elsif(($modifiers[$i] =~ m/^optional_arg_is_void_ptr$/) &&
+			($fn->{'has_varargs'})) {
+			$r_param_names->[$varargs_index] = "opt_arg";
+			$r_param_types->[$varargs_index] = "void *";
+			$mods->{'local_vars_for_varargs_handler'} .=
+				"\tvoid *opt_arg = NULL;\n";
+			$mods->{'va_list_handler_code'} =
+				create_code_for_va_list_get_void_ptr(
+					$fn->{'last_named_var'});
+			$varargs_handled = 1;
 		} elsif(($modifiers[$i] eq 'pass_va_list') &&
 			($fn->{'has_varargs'}) &&
 			($command eq 'GATE')) {
 
 			$r_param_names->[$varargs_index] = "ap";
+			$r_param_types->[$varargs_index] = "va_list";
 			$mods->{'local_vars_for_varargs_handler'} .=
 				"\tva_list ap;\n";
 			$mods->{'va_list_handler_code'} = "\tva_start(ap,".
@@ -562,6 +645,68 @@ sub process_wrap_or_gate_modifiers {
 	return($mods);
 }
 
+sub create_postprocessors {
+	my $fn = shift;
+	my $mods = shift;
+
+	my $num_params_to_postprocess = @{$mods->{'postprocess_vars'}};
+	my $postprocessor_calls = undef;
+	my $postprocessor_prototypes = "";
+	my $fn_name = $fn->{'fn_name'};
+
+	my $return_value_param_in_call = "";
+	my $return_value_param_in_prototype = "";
+
+	my $fn_return_type = $fn->{'fn_return_type'};
+	if($fn_return_type ne "void") {
+		$return_value_param_in_call = "ret, ";
+		$return_value_param_in_prototype = $fn_return_type." ret, ";
+	}
+
+	if ($num_params_to_postprocess > 0) {
+		$postprocessor_calls = "";
+
+		# insert call to postprocessor for each variable
+		my $ppvar;
+		foreach $ppvar (@{$mods->{'postprocess_vars'}}) {
+			my $pp_fn = "${fn_name}_postprocess_${ppvar}";
+
+			if ($ppvar eq '') {
+				# postprocess, no mapped parameter
+				$postprocessor_calls .= "$pp_fn(__func__, ".
+					$return_value_param_in_call.
+					# add orig (unmapped) parameters
+					join(", ", @{$mods->{'parameter_names'}}).
+					"); ";
+				$postprocessor_prototypes .= "extern void ".
+					"$pp_fn(const char *realfnname, ".
+					$return_value_param_in_prototype.
+					# orig (unmapped) parameters
+					join(", ", @{$mods->{'parameter_types'}}).
+					");\n";
+			} else {
+				# has mapped parameter
+				my $mapping_results = $mods->{'mapping_results_by_orig_name'}->{$ppvar};
+
+				$postprocessor_calls .= "$pp_fn(__func__, ".
+					$return_value_param_in_call.
+					"&$mapping_results, ".
+					# add orig (unmapped) parameters
+					join(", ", @{$mods->{'parameter_names'}}).
+					"); ";
+				$postprocessor_prototypes .= "extern void ".
+					"$pp_fn(const char *realfnname, ".
+					$return_value_param_in_prototype.
+					"mapping_results_t *res, ".
+					# orig (unmapped) parameters
+					join(", ", @{$mods->{'parameter_types'}}).
+					");\n";
+			}
+		}
+	}
+	return($postprocessor_calls, $postprocessor_prototypes);
+}
+
 sub create_call_to_real_fn {
 	my $fn = shift;
 	my $mods = shift;
@@ -569,6 +714,11 @@ sub create_call_to_real_fn {
 
 	my $real_fn_pointer_name = $mods->{'real_fn_pointer_name'};
 
+	my $postprocessor_calls = undef;
+	my $postprocessor_prototypes = "";
+	($postprocessor_calls, $postprocessor_prototypes) = 
+		create_postprocessors($fn, $mods);
+
 	return(
 		# 1. call with mapped parameters
 		"(*$real_fn_pointer_name)(".
@@ -578,7 +728,11 @@ sub create_call_to_real_fn {
 			join(", ", @{$mods->{'parameter_names'}}).");\n",
 		# 3. call with original parameters (without logging)
 		"(*$real_fn_pointer_name)(".
-			join(", ", @{$mods->{'parameter_names'}}).");\n"
+			join(", ", @{$mods->{'parameter_names'}}).");\n",
+		# 4.
+		$postprocessor_calls,
+		# 5. prototypes.
+		$postprocessor_prototypes
 	);
 }
 
@@ -627,14 +781,26 @@ sub create_call_to_gate_fn {
 	}
 
 	my $mapped_call = "${fn_name}_gate($modified_param_list);\n";
+
+	my $postprocessor_calls = undef;
+	my $postprocessor_prototypes = "";
+	($postprocessor_calls, $postprocessor_prototypes) = 
+		create_postprocessors($fn, $mods);
+
 	my $unmapped_call = "${fn_name}_gate($orig_param_list);\n";
-	# nomap_nolog is not possible for GATEs
+
+	# nomap_nolog for a gate is a direct call to the real function
+	my $unmapped_nolog_call = "${fn_name}_next__(".
+		join(", ", @{$mods->{'parameter_names'}}).
+		");\n";
 
 	my $gate_function_prototype =
 		"extern ".$fn->{'fn_return_type'}." ${fn_name}_gate(".
 			"$prototype_params);\n";
 
-	return($mapped_call, $unmapped_call, $gate_function_prototype);
+	return($mapped_call, $unmapped_call, $unmapped_nolog_call,
+		$gate_function_prototype.$postprocessor_prototypes,
+		$postprocessor_calls);
 }
 
 #-------------------
@@ -679,6 +845,8 @@ my $export_h_buffer =
 #include <sys/xattr.h>
 #endif
 
+#include \"mapping.h\"
+
 ";
 
 # Handle "WRAP" and "GATE" commands.
@@ -743,24 +911,52 @@ sub command_wrap_or_gate {
 		$nomap_fn_c_code .=	"\t$fn_return_type ret;\n";
 		$nomap_nolog_fn_c_code .= "\t$fn_return_type ret;\n";
 	}
-	$wrapper_fn_c_code .=	"\tint saved_errno = errno;\n";
+	$wrapper_fn_c_code .=	"\tint saved_errno = errno;\n".
+				"\terrno = 0;\n";
 	$nomap_fn_c_code .=	"\tint saved_errno = errno;\n";
 
+	# variables have been introduced, add the code:
+	$wrapper_fn_c_code .=		$common_initcode_for_all_functions;
+	$nomap_fn_c_code .=		$common_initcode_for_all_functions;
+	$nomap_nolog_fn_c_code .=	$common_initcode_for_all_functions;
+
 	$wrapper_fn_c_code .=		$mods->{'path_mapping_code'}.
 					$mods->{'path_ro_check_code'};
 	$wrapper_fn_c_code .=		$mods->{'va_list_handler_code'};
 	$nomap_fn_c_code .=		$mods->{'va_list_handler_code'};
 	$nomap_nolog_fn_c_code .=	$mods->{'va_list_handler_code'};
 
+	my $loglevel_no_real_fn;
+	my $no_real_fn_abort_code;
+	if($command eq 'WRAP') {
+		# Wrappers log an error and abort if the real function
+		# does not exist.
+		$loglevel_no_real_fn = "SB_LOGLEVEL_ERROR";
+		$no_real_fn_abort_code = "abort();";
+	} else { # GATE
+		# Gates log a warning (but don't abort) if the real function
+		# does not exist - the gate function should handle the rest.
+		$loglevel_no_real_fn = "SB_LOGLEVEL_WARNING";
+		$no_real_fn_abort_code = "/* no abort() */";
+	}
+
 	my $check_fn_pointer_log_enabled .=
 		"\tif($real_fn_pointer_name == NULL) {\n".
 		"\t\t$real_fn_pointer_name = sbox_find_next_symbol(1, ".
 			"\"$fn_name\");\n".
+		"\t\tif ($real_fn_pointer_name == NULL) {\n".
+		"\t\t\tSB_LOG($loglevel_no_real_fn, \"Real '%s'".
+			" not found\", \"$fn_name\");\n".
+		"\t\t\t$no_real_fn_abort_code\n".
+		"\t\t}\n".
 		"\t}\n";
 	my $check_fn_pointer_log_disabled .=
 		"\tif($real_fn_pointer_name == NULL) {\n".
 		"\t\t$real_fn_pointer_name = sbox_find_next_symbol(0, ".
 			"\"$fn_name\");\n".
+		"\t\tif ($real_fn_pointer_name == NULL) {\n".
+		"\t\t\t$no_real_fn_abort_code\n".
+		"\t\t}\n".
 		"\t}\n";
 	$wrapper_fn_c_code .=		$check_fn_pointer_log_enabled;
 	$nomap_fn_c_code .=		$check_fn_pointer_log_enabled;
@@ -783,9 +979,12 @@ sub command_wrap_or_gate {
 	my $call_line_prefix = "\t";
 	my $return_statement = ""; # return stmt not needed if fn_type==void
 	my $log_return_val = ""; 
+	if($mods->{'return_expr'}) {
+		$return_statement .= "\tret = ".$mods->{'return_expr'}.";\n";
+	}
 	if($fn_return_type ne "void") {
 		$call_line_prefix .= "ret = ";
-		$return_statement = "\treturn(ret);\n";
+		$return_statement .= "\treturn(ret);\n";
 	}
 
 	# create a call to log the return
@@ -832,22 +1031,35 @@ sub command_wrap_or_gate {
 	my $mapped_call;
 	my $unmapped_call;
 	my $unmapped_nolog_call;
+	my $postprocesors;
+	my $prototypes;
 	if($command eq 'WRAP') {
-		($mapped_call, $unmapped_call, $unmapped_nolog_call) =
-			create_call_to_real_fn($fn, $mods,
-				@param_list_in_next_call);
+		($mapped_call, $unmapped_call, $unmapped_nolog_call,
+		 $postprocesors, $prototypes) = create_call_to_real_fn(
+			$fn, $mods, @param_list_in_next_call);
 	} else { # GATE
-		my $gate_function_prototype;
-		($mapped_call, $unmapped_call, $gate_function_prototype) =
-			create_call_to_gate_fn($fn, $mods,
-				@param_list_in_next_call);
-		$export_h_buffer .= $gate_function_prototype;
-		$unmapped_nolog_call = ""; # not supported for GATEs
+		($mapped_call, $unmapped_call, $unmapped_nolog_call,
+		 $prototypes, $postprocesors) = create_call_to_gate_fn(
+			$fn, $mods, @param_list_in_next_call);
 	}
+	$export_h_buffer .= $prototypes;
+
+	# First restore errno to what it was at entry (the path mapping
+	# code might have set it)
+	$wrapper_fn_c_code .=		"\terrno = saved_errno;\n";
+	$nomap_fn_c_code .=		"\terrno = saved_errno;\n";
+
+	# Next, insert the call to the real function
 	$wrapper_fn_c_code .=		$call_line_prefix.$mapped_call;
 	$nomap_fn_c_code .=		$call_line_prefix.$unmapped_call;
 	$nomap_nolog_fn_c_code .=	$call_line_prefix.$unmapped_nolog_call;
 
+	# calls to postprocessors (if any) before the cleanup 
+	if (defined $postprocesors) {
+		$wrapper_fn_c_code .=	"\t".$postprocesors."\n";
+	}
+	
+	# cleanup; free allocated variables etc.
 	$wrapper_fn_c_code .=		$mods->{'va_list_end_code'};
 	$wrapper_fn_c_code .=		$mods->{'free_path_mapping_vars_code'};
 	$nomap_fn_c_code .=		$mods->{'va_list_end_code'};
@@ -939,7 +1151,7 @@ while ($line = <STDIN>) {
 	$line =~ s/\s$//;
 
 	# Split to fields. 1st=command, 2nd=function, 3rd=modifiers
-	my @field = split(/\s*:\s*/, $line);
+	my @field = split(/\s*:\s*/, $line, 3);
 
 	# Order the cached pre-processor token first
 	if ($token_cache) {
